(*---------------------------------------------------------------------
 *     		A toy implementation of the
 *	 	
 *		I m p e r a t i v e  R h o - c a l c u l u s
 *
 *     		using Coq
 *
 *    Copyright (c) 2003 Bernard Paul Serpette and Luigi Liquori
 *
 *      	Bug descriptions, use reports,
 *	comments or suggestions are
 *      	welcome. Send them to
 *	
 *	Bernard.Serpette@inria.fr
 *	Luigi.Liquori@inria.fr
 *
 *    This program is free software * you can redistribute it
 *    and/or modify it under the terms of the GNU General Public
 *    License as published by the Free Software Foundation * either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY * without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public
 *    License along with this program * if not, write to the Free
 *    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *    MA 02111-1307, USA.
 *--------------------------------------------------------------------- *)
Require Export TheoryList.
Require Export Arith.
Require Export EqNat.
Require Export Bool.

Section TypedImpRho.

Implicit Arguments On.

Axiom Skip : False.
Tactic Definition Damn := Cut(False);[(Intro Fa; Inversion Fa)|Apply Skip].

Tactic Definition Push IT :=
 Generalize IT; Clear IT; Intros IT.

Tactic Definition Switch PRED :=
 Cut(EX b:bool | b=PRED); [Intros (BBBBB,RES) | Exists PRED]; Auto;
 Generalize RES;
 Case PRED; Intros FFFFF; Rewrite FFFFF in RES; Clear FFFFF BBBBB.

(*
 *
 * Partial Functions (Library)
 *
 *)
Definition PartialFunction[A:Set][B:Set] := A -> (option B).

Definition Eq[A:Set][B:Set] :
 (PartialFunction A B) -> (PartialFunction A B) -> Prop :=
 [f:(PartialFunction A B)][g:(PartialFunction A B)]
 ((x:A) (f x)=(g x)).

Lemma Eq_refl: (A:Set)(B:Set)(f:(PartialFunction A B))
 (Eq f f).
Proof.
Unfold Eq; Auto.
Qed.
Hints Resolve Eq_refl.

Lemma Eq_sym: (A:Set)(B:Set)(f,g:(PartialFunction A B))
 (Eq f g) -> (Eq g f).
Proof.
Unfold Eq; Auto.
Qed.
Hints Resolve Eq_sym.

Definition Extend[A:Set][Eq:A -> A -> bool][B:Set] :
 (PartialFunction A B) -> A -> B -> (PartialFunction A B)
 := [f:(PartialFunction A B)][x1:A][y:B][x2:A]
 (if (Eq x1 x2) then (Some B y) else (f x2)).

Definition InDomain[A:Set][B:Set] :
 A -> (PartialFunction A B) -> Prop
 := [x:A][f:(PartialFunction A B)]
 (EX y:B | (f x)=(Some B y)).

Definition LowerDomain[A:Set][B:Set] :
 (PartialFunction A B) -> (PartialFunction A B) -> Prop :=
 [f:(PartialFunction A B)][g:(PartialFunction A B)]
 ((x:A)(y:B) (f x)=(Some B y) -> (EX z:B | (g x)=(Some B z))).

Lemma LowerDomain_refl: (A:Set)(B:Set)(f:(PartialFunction A B))
 (LowerDomain f f).
Proof.
Unfold LowerDomain; Intros; Exists y; Auto.
Qed.
Hints Resolve LowerDomain_refl.

Lemma LowerDomain_trans: (A:Set)(B:Set)(f1,f2:(PartialFunction A B))
 (LowerDomain f1 f2) -> (f3:(PartialFunction A B))
 (LowerDomain f2 f3) ->
 (LowerDomain f1 f3).
Proof.
Unfold LowerDomain; Intros.
LApply (H x y); Auto; Intros (z,F2); EAuto.
Qed.
Hints Resolve LowerDomain_trans.

Definition Lower[A:Set][B:Set] :
 (PartialFunction A B) -> (PartialFunction A B) -> Prop :=
 [f:(PartialFunction A B)][g:(PartialFunction A B)]
 ((x:A)(y:B) (f x)=(Some B y) -> (g x)=(Some B y)).

Lemma Lower_better_LowerDomain: (A:Set)(B:Set)(f,g:(PartialFunction A B))
 (Lower f g) -> (LowerDomain f g).
Proof.
Unfold Lower LowerDomain; Intros; EAuto.
Qed.
Hints Resolve Lower_better_LowerDomain.

Lemma Lower_refl: (A:Set)(B:Set)(f:(PartialFunction A B)) (Lower f f).
Proof.
Unfold Lower; Auto.
Qed.
Hints Resolve Lower_refl.

Lemma Lower_trans: (A:Set)(B:Set)(f1,f2:(PartialFunction A B))
 (Lower f1 f2) -> (f3:(PartialFunction A B))
 (Lower f2 f3) ->
 (Lower f1 f3).
Proof.
Unfold Lower; Auto.
Qed.
Hints Resolve Lower_trans.

Lemma Fresh_LowerExtend:
   (A:Set)(Eq:A -> A -> bool)(RealEq: ((a,b:A) true=(Eq a b) -> (a=b)))(B:Set)
   (f:(PartialFunction A B))(x:A)
 (f x)=(None B) -> (y:B)
 (Lower f (Extend Eq f x y)).
Proof.
Unfold Lower Extend; Intros A EQ REQ B f x NONE y x0 y0 GET.
Switch '(EQ x x0); Auto.
LApply (REQ x x0); Auto.
Intros EQx; Rewrite EQx in NONE; Rewrite NONE in GET; Inversion GET.
Qed.
Hints Resolve Fresh_LowerExtend.

(*
 *
 * Domains
 *
 *)
Variable basic:Set.
Variable eqbasic : basic -> basic -> bool.
Variable var:Set.
Variable eqvar : var -> var -> bool.
Variable boperator:Set.

Definition loc := nat.

(* Type *)
Inductive type : Set :=
   Basic: basic -> type
 | FunType: type -> type -> type
 | ProdType: type -> type -> type
 | RefType: type -> type.

Definition operator := boperator*type.

(* Pattern *)
Inductive pattern : Set :=
   POp: operator -> (list pattern) -> pattern
 | PVar: var -> type -> pattern
 | PCons: pattern -> pattern -> pattern
 | PRef: pattern -> pattern.

Definition patterns := (list pattern).

(* Expression *)
Inductive expr : Set :=
   Op: operator -> expr
 | Var: var -> expr
 | Abs: pattern -> expr -> expr
 | App: expr -> expr -> expr
 | Cons: expr -> expr -> expr
 | Ref: expr -> expr
 | Deref: expr -> expr
 | Assign: expr -> expr -> expr.

(* Value *)
Inductive value : Set :=
   VOp: operator -> (list value) -> value
 | Pair: value -> value -> value
 | Closure: pattern -> expr -> (var -> (option value)) -> value
 | Wrong: pattern -> value -> expr -> (var->(option value)) -> value
 | Loc: loc -> value.

Definition env := (PartialFunction var value).
Definition extend_env := (Extend eqvar).
Definition store := (PartialFunction loc value).
Definition extend_store := (Extend beq_nat).
Definition envt := (PartialFunction var type).
Definition extend_envt := (Extend eqvar).
Definition storet := (PartialFunction loc type).
Definition extend_storet := (Extend beq_nat).

Definition values := (list value).

Definition env_init : var -> (option value) := [x:var] (None value).
Definition envt_init : var -> (option type) := [x:var] (None type).
Definition store_init : loc -> (option value) := [i:loc] (None value).
Definition storet_init : loc -> (option type) := [i:loc] (None type).

Variable sbrk: store -> loc.
Variable sbrk_valid: (s:store)(i:loc) (i=(sbrk s)) -> (s i)=(None value).

(********** Writable store for sbrk **********)
Lemma LowerWhenSbrk: (s:store)(i:loc)
 i=(sbrk s) -> (a:value)
 (Lower s (extend_store s i a)).
Proof.
Intros s i SBRK a.
Cut((s i)=(None value)); [Intros FRESH | Apply sbrk_valid]; Auto.
Apply(Fresh_LowerExtend beq_nat_eq FRESH).
Qed.
Hints Resolve LowerWhenSbrk.

(********** Writable store **********)
Lemma LowerDomainWhenWrite: (s:store)(i:loc)(x:value)
 (s i)=(Some value x) -> (y:value)
 (LowerDomain s (extend_store s i y)).
Proof.
Unfold LowerDomain extend_store; Unfold Extend; Intros.
Switch '(beq_nat i x0); [Exists y | Exists y0]; Auto.
Qed.
Hints Resolve LowerDomainWhenWrite.

(*
 * The Matching Rules
 *)
Inductive match : pattern -> value -> store -> env -> env -> Prop :=
   matchOpNil: (op:operator)(s:store)(e:env)
	(match (POp op (nil ?)) (VOp op (nil ?)) s e e)
 | matchOpCons: (op:operator)(lp:patterns)(l:values)(s:store)(e,e':env)
	(match (POp op lp) (VOp op l) s e e') ->
        (p:pattern)(v:value)(r:env)
	(match p v s e' r) ->
	(match (POp op (cons p lp)) (VOp op (cons v l)) s e r)
 | matchVar:(v:var)(t:type)(val:value)(s:store)(e:env)
	(match (PVar v t) val s e (extend_env e v val))
 | matchCons: (left:pattern)(car:value)(s:store)(e,e':env)
	(match left car s e e') -> (right:pattern)(cdr:value)(r:env)
	(match right cdr s e' r) ->
	(match (PCons left right) (Pair car cdr) s e r)
 | matchRef: (i:loc)(s:store)(v:value)
	((s i)=(Some value v)) -> (x:pattern)(e,r:env)
	(match x v s e r) ->
	(match (PRef x) (Loc i) s e r).
Hints Resolve matchOpNil matchOpCons matchVar matchCons matchRef.

(********** Determinism **********)
Lemma match_fun : (P:pattern)(v:value)(s:store)(e:env)(e1:env)
 (match P v s e e1) -> (e2:env)
 (match P v s e e2) ->
 (e1=e2).
Proof.
Intros P v s e e1 M1; Elim M1; Clear M1 e1 e s v P.
 (* Op *)
 Intros op s e e2 M; Inversion M; Auto.
 (* Op *)
 Intros op lp l s e e' Ml rMl p v r Mp rMp e2 M; Inversion M; Auto.
 Generalize H8; LApply (rMl e'0); Auto; Intros EQE.
 Rewrite <- EQE; Intros MA2.
 Apply (rMp e2); Auto.
 (* Var *)
 Intros v t val s e e2 M; Inversion M; Auto.
 (* Cons *)
 Intros left car s e e' ML rML right cdr r MR rMR e2 M; Inversion M; Auto.
 Generalize H7; LApply (rML e'0); Auto; Intros EQE.
 Rewrite <- EQE; Intros MR2.
 Apply (rMR e2); Auto.
 (* Ref *)
 Intros i s v READ x e r MX rMX e2 M; Inversion M; Auto.
 Generalize H2; Rewrite H1 in READ; Inversion READ; Intros M2; Auto.
Qed.
Hints Resolve match_fun.

(********** Coherence of lists length in ?Op **********)
Lemma Match_Op_Length : (op:operator)(lp:patterns)(l:values)(s:store)(e,e':env)
 (match (POp op lp) (VOp op l) s e e') ->
 ((length lp)=(length l)).
Proof.
Intros op lp; Elim lp.
 Intros l s e e' M; Inversion M; Auto.
Intros a l; Elim l.
 Intros R l0 s e e' M; Inversion M.
 LApply (R l1 s e e'0); Auto; Clear R; Unfold length; Intros EQLen.
 Inversion EQLen; Auto.
Intros a0 l0 BigR R l1 s e e' M; Inversion M.
LApply (R l2 s e e'0); Auto; Intros EQLen.
Cut((S (length (cons a0 l0)))=(S (length l2)));[Intros | Auto].
Unfold length; Auto.
Qed.
Hints Resolve Match_Op_Length.

(*
 * Evaluator
 *)
Mutual Inductive eval : expr -> env -> store -> value -> store -> Prop :=
   evalOp: (o:operator)(e:env)(s:store)
	(eval (Op o) e s (VOp o (nil value)) s)
 | evalVar: (x:var)(e:env)(v:value)
	((e x)=(Some value v)) -> (s:store)
	(eval (Var x) e s v s)
 | evalAbs: (P:pattern)(B:expr)(e:env)(s:store)
	(eval (Abs P B) e s (Closure P B e) s)
 | evalApp: (F:expr)(e:env)(s:store)(f:value)(s1:store)
	(eval F e s f s1) -> (A:expr)(a:value)(s2:store)
	(eval A e s1 a s2) -> (v:value)(s3:store)
	(call f a s2 v s3) ->
	(eval (App F A) e s v s3)
 | evalCons: (A:expr)(e:env)(s:store)(a:value)(s1:store)
	(eval A e s a s1) -> (B:expr)(b:value)(s2:store)
	(eval B e s1 b s2) ->
	(eval (Cons A B) e s (Pair a b) s2)
 | evalRef: (A:expr)(e:env)(s:store)(a:value)(s1:store)
	(eval A e s a s1) -> (i:loc)
	(i=(sbrk s1)) ->
	(eval (Ref A) e s (Loc i) (extend_store s1 i a))
 | evalDeref: (A:expr)(e:env)(s:store)(i:loc)(s1:store)
	(eval A e s (Loc i) s1) -> (v:value)
	((s1 i)=(Some value v)) ->
	(eval (Deref A) e s v s1)
 | evalAssign: (A:expr)(e:env)(s:store)(i:loc)(v:value)(s1:store)
	(eval A e s (Loc i) s1) -> (B:expr)(s2:store)
	(eval B e s1 v s2) -> (old:value) (* force *)
	((s1 i)=(Some value old)) ->
	(eval (Assign A B) e s v (extend_store s2 i v))

(*
 * Caller
 *)
with      call : value -> value -> store -> value -> store -> Prop :=
 | callVOp: (op:operator)(l:values)(v:value)(s:store)
	(call (VOp op l) v s (VOp op (cons v l)) s)
 | callClosureOK: (P:pattern)(v:value)(s:store)(e,e':env)
	(match P v s e e') -> (B:expr)(r:value)(s1:store)
	(eval B e' s r s1) ->
	(call (Closure P B e) v s r s1)
 | callClosureKO: (P:pattern)(v:value)(s:store)(e:env)
	(not (EX e':env | (match P v s e e'))) -> (B:expr)
	(call (Closure P B e) v s (Wrong P v B e) s)
 | callPair: (left:value)(v:value)(s:store)(r1:value)(s1:store)
	(call left v s r1 s1) -> (right:value)(r2:value)(s2:store)
	(call right v s1 r2 s2) ->
	(call (Pair left right) v s (Pair r1 r2) s2).
Hints Resolve evalVar evalOp evalAbs evalApp evalCons evalRef.
Hints Resolve evalDeref evalAssign.
Hints Resolve callVOp callClosureOK callClosureKO callPair.

Scheme neval_ind := Minimality for eval Sort Prop
  with ncall_ind := Minimality for call Sort Prop.

(********** No Garbage Collection : the store grows **********)
Definition NGC_eval : expr -> env -> store -> value -> store -> Prop :=
 [E:expr][e:env][s:store][v:value][s1:store]
 (LowerDomain s s1).

Definition NGC_call : value -> value -> store -> value -> store -> Prop :=
 [f:value][a:value][s:store][v:value][s1:store]
 (LowerDomain s s1).

Lemma NoGarbageCollection: (E:expr)(e:env)(s:store)(v:value)(s1:store)
 (eval E e s v s1) -> (LowerDomain s s1).
Proof.
Intros E e s v s1 EV;
Elim EV using neval_ind with NGC_eval NGC_call E e s v s1;
Clear EV E e s v s1; Unfold NGC_eval NGC_call; Auto.
(* Only transitivity is needed *) EAuto. EAuto.
(* Ref *)
Intros.
Cut(Lower s1 (extend_store s1 i a)); [Intros LE | Auto]; Auto.
Cut(LowerDomain s1 (extend_store s1 i a)); [EAuto | EAuto]; Auto.
(* Assign *)
Intros.
Cut(LowerDomain s s2); [Intros LD2 | EAuto].
Unfold LowerDomain in H2.
LApply (H2 i old); Auto; Intros (z,S2).
Cut(LowerDomain s2 (extend_store s2 i v)); [Intros LD3 | EAuto]; EAuto.
(* Only transitivity is needed *) EAuto.
Qed.
Hints Resolve NoGarbageCollection.

(********** eval is a (partial function) **********)
Definition the_P : expr -> env -> store -> value -> store -> Prop :=
 [E:expr][e:env][s:store][v:value][s':store]
  ((v2:value)(s2:store) (eval E e s v2 s2) -> ((v=v2) /\ (s'=s2))).

Definition the_Q : value -> value -> store -> value -> store -> Prop :=
 [f:value][a:value][s:store][v:value][s':store]
  ((v2:value)(s2:store) (call f a s v2 s2) -> ((v=v2) /\ (s'=s2))).

Theorem eval_deterministic: (A:expr)(e:env)(s:store)(v1:value)(s1:store)
 (eval A e s v1 s1) -> (v2:value)(s2:store)
 (eval A e s v2 s2) ->
 ((v1=v2) /\ (s1=s2)).
Proof.
Intros A e s v1 s1 EV1;
Elim EV1 using neval_ind with the_P the_Q A e s v1 s1;
Clear EV1 s1 v1 s e A; Unfold the_P; Unfold the_Q; EAuto.
 (* Op *)
 Intros o e s v2 s2 EV; Inversion EV; Auto.
 (* Var *)
 Intros x e v GET s v2 s2 EV; Inversion EV; Split; Auto.
 Rewrite H0 in GET; Inversion GET; Auto.
 (* Abs *)
 Intros P B e s v2 s2 EV; Inversion EV; Auto.
 (* App *)
 Intros F e s f s1 EVF rF A a s2 EVA rA v s3 CALL rCALL v2 s4 EV; Inversion EV.
 Generalize H2 H7; LApply (rF f0 s5); Auto; Clear EVF rF.
 Intros (EQF,EQS); Rewrite <- EQF; Rewrite <- EQS; Intros EVA2.
 LApply (rA a0 s6); Auto; Clear EVA EVA2 rA.
 Intros (EQF2,EQS2); Rewrite <- EQF2; Rewrite <- EQS2; Intros CALL2.
 LApply (rCALL v2 s4); Auto.
 (* Cons *)
 Intros A e s a s1 EVA recA B b s2 EVB recB v2 s3 EV2; Inversion EV2.
 Generalize H6; LApply (recA a0 s4); Auto.
 Intros (EQA,EQS); Rewrite <- EQA; Rewrite <- EQS; Intros EVB2.
 LApply (recB b0 s3); Auto.
 Intros (EQB,EQS2); Rewrite <- EQB; Rewrite <- EQS2; Auto.
 (* Ref *)
 Intros A e s a s1 EVA recA n LEN v2 s2 EV2; Inversion EV2.
 Generalize H1; LApply (recA a0 s3); Auto.
 Intros (EQA,EQS); Rewrite <- EQA; Rewrite <- EQS; Intros LEN2.
 Rewrite <- LEN in LEN2; Rewrite LEN2; Auto.
 (* Deref *)
 Intros A e s n s1 EVA recA v GETS v2 s2 EV2; Inversion EV2.
 Generalize H1; LApply (recA (Loc i) s2); Auto.
 Intros (EQL,EQS); Inversion EQL; Rewrite <- EQS; Rewrite <- H7; Intros GETS2.
 Rewrite GETS2 in GETS; Inversion GETS; Auto.
 (* Assign *)
 Intros A e s i v s1 EVA rA B s2 EVB rB old READ v2 s3 EV2; Inversion EV2.
 Generalize H2; LApply (rA (Loc i0) s4); Auto.
 Intros (EQV,EQS); Inversion EQV; Rewrite <- EQS; Intro EVB2.
 LApply (rB v2 s5); Auto.
 Intros (EQL,EQS2); Rewrite <- EQL; Rewrite <- EQS2; Auto.
 (* CALL *)
 (* VOp *)
 Intros op l v s v2 s2 CALL; Inversion CALL; Auto.
 (* ClosureOK *)
 Intros P v s e e' MATCH B r s1 EVB recB v2 s2 CALL; Inversion CALL; Auto.
  (* OK OK *)
  Generalize H7; Cut(e'=e'0); [Intros EQE | Apply (match_fun MATCH H6)].
  Rewrite <- EQE; Intros EVB2; LApply (recB v2 s2); Auto.
  (* OK KO *)
  Elim H6; Rewrite <- H5; Exists e'; Auto.
 (* ClosureKO *)
 Intros P v s e NOTEX B v2 s2 CALL; Inversion CALL; Auto. (* KO OK *)
 Elim NOTEX; Exists e'; Auto.
 (* Pair *)
 Intros left v s r1 s1 CALLL rCALLL right r2 s2 CALLR rCALLR v2 s3 CALL;
 Inversion CALL.
 Generalize H6; LApply (rCALLL r0 s4); Auto.
 Intros (EQR,EQS); Rewrite <- EQR; Rewrite <- EQS; Intros CALLR2.
 LApply (rCALLR r3 s3); Auto.
 Intros (EQR2,EQS2); Rewrite <- EQR2; Rewrite <- EQS2; Auto.
Qed.
Hints Resolve eval_deterministic.

(********** Determinism of the caller **********)
Theorem call_deterministic: (v1,v2:value)(s:store)(r1:value)(s1:store)
 (call v1 v2 s r1 s1) -> (r2:value)(s2:store)
 (call v1 v2 s r2 s2) ->
 ((r1=r2) /\ (s1=s2)).
Proof.
(* This is bureaucratia since we cut and paste the end of the previous
   proof and substitute Rec Hypothetis by deterministic itself.
   How we can automatize that. Please Help Doctor Yves.
*)
Intros v1 v2 s r1 s1 CALL; Elim CALL; Clear CALL s1 r1 s v2 v1; Auto.
 (* VOp *)
 Intros op l v s v2 s2 CALL; Inversion CALL; Auto.
 (* ClosureOK *)
 Intros P v s e e' MATCH B r s1 EVB v2 s2 CALL; Inversion CALL; Auto.
  (* OK OK *)
  Generalize H7; Cut(e'=e'0); [Intros EQE | EAuto].
  Rewrite <- EQE; Intros EVB2; EAuto.
  (* OK KO *)
  Elim H6; Rewrite <- H5; Exists e'; Auto.
 (* ClosureKO *)
 Intros P v s e NOTEX B v2 s2 CALL; Inversion CALL; Auto. (* KO OK *)
 Elim NOTEX; Exists e'; Auto.
 (* Pair *)
 Intros left v s r1 s1 CALLL rCALLL right r2 s2 CALLR rCALLR v2 s3 CALL;
 Inversion CALL.
 Generalize H6; LApply (rCALLL r0 s4); Auto.
 Intros (EQR,EQS); Rewrite <- EQR; Rewrite <- EQS; Intros CALLR2.
 LApply (rCALLR r3 s3); Auto.
 Intros (EQR2,EQS2); Rewrite <- EQR2; Rewrite <- EQS2; Auto.
Qed.
Hints Resolve call_deterministic.

(*
 * Coercion to a functional type
 *)
Inductive NormalizeFunType : type -> type -> Prop :=
   nftRefl: (t1,t2:type)
	(NormalizeFunType (FunType t1 t2) (FunType t1 t2))
 | nftProd: (t1,t3,t4:type)
	(NormalizeFunType t1 (FunType t3 t4)) -> (t2,t5:type)
	(NormalizeFunType t2 (FunType t3 t5)) ->
	(NormalizeFunType (ProdType t1 t2) (FunType t3 (ProdType t4 t5))).
Hints Resolve nftRefl nftProd.

(********** The same functional definition **********)
Fixpoint eqType [t1,t2:type] : bool :=
 Cases t1 t2 of
 | (Basic b) (Basic b1) => (eqbasic b b1)
 | (FunType t1 t2) (FunType t3 t4) =>
	(andb (eqType t1 t3) (eqType t2 t4))
 | (ProdType t1 t2) (ProdType t3 t4) =>
	(andb (eqType t1 t3) (eqType t2 t4))
 | (RefType t1) (RefType t2) =>
 	 (eqType t1 t2)
 | _ _ => false
end.

Fixpoint CoerceToFunType [t:type] : (option type) :=
 Cases t of
 | (FunType t1 t2) => (Some type t)
 | (ProdType t1 t2) =>
	Cases (CoerceToFunType t1) (CoerceToFunType t2) of
	| (Some (FunType t3 t4)) (Some (FunType t3' t5)) =>
		if (eqType t3 t3')
			then (Some type (FunType t3 (ProdType t4 t5)))
			else (None type)
	| _ _ => (None type)
	end
 | _ => (None type)
 end.

(********** NormalizeFunType is a (partial) function **********)
Lemma NormalizeFunType_deterministic: (t,t1:type)
 (NormalizeFunType t t1) -> (t2:type)
 (NormalizeFunType t t2) ->
 (t1=t2).
Proof.
Intros t t1 NFT; Induction NFT; Intros; Auto; Inversion H; Auto.
LApply (HrecNFT1 (FunType t7 t8)); Auto; Intros FF; Inversion FF.
LApply (HrecNFT0 (FunType t7 t10)); Auto; Intros GG; Inversion GG.
Auto.
Qed.
Hints Resolve NormalizeFunType_deterministic.

(*
 * SubTyping (Not yet used)
 *)
Inductive SubType : type -> type -> Prop :=
   stBasic: (b:basic)
	(SubType (Basic b) (Basic b))
 | stStruct: (t1,t3:type)
	(SubType t1 t3) -> (t2,t4:type)
	(SubType t2 t4) ->
	(SubType (ProdType t1 t2) (ProdType t3 t4))
 | stProj1: (t1,t2:type)
	(SubType (ProdType t1 t2) t1)
 | stProj2: (t1,t2:type)
	(SubType (ProdType t1 t2) t2)
 | stArrow: (t1,t3:type)
	(SubType t3 t1) -> (t2,t4:type)
	(SubType t2 t4) ->
	(SubType (FunType t1 t2) (FunType t3 t4))
 | stDeref: (t1,t2:type)
	(SubType t1 t2) ->
	(SubType (RefType t1) (RefType t2)).
Hints Resolve stBasic stStruct stProj1 stProj2 stArrow stDeref.

(*
 * Type Check Pattern
 *)
Inductive TypeCheckPattern : envt -> pattern -> envt -> type -> Prop :=
 | tcPOpNil: (E:envt)(op:boperator)(t:type)
	(TypeCheckPattern E (POp (op,t) (nil ?)) E t)
 | tcPOpCons: (E,E1:envt)(op:operator)(lp:patterns)(t:type)
	(TypeCheckPattern E (POp op lp) E1 t)  -> (t1,t2:type)
	(NormalizeFunType t (FunType t1 t2)) -> (P:pattern)(E2:envt)
	(TypeCheckPattern E1 P E2 t1) ->
	(TypeCheckPattern E (POp op (cons P lp)) E2 t2)
 | tcPVar: (E:envt)(x:var)(t:type)
	(TypeCheckPattern E (PVar x t) (extend_envt E x t) t)
 | tcPCons: (E,E1:envt)(A:pattern)(t1:type)
	(TypeCheckPattern E A E1 t1) -> (B:pattern)(t2:type)(E2:envt)
	(TypeCheckPattern E1 B E2 t2) ->
	(TypeCheckPattern E (PCons A B) E2 (ProdType t1 t2))
 | tcPRef: (E,E1:envt)(A:pattern)(t:type)
	(TypeCheckPattern E A E1 t) ->
	(TypeCheckPattern E (PRef A) E1 (RefType t)).
Hints Resolve tcPOpNil tcPOpCons tcPVar tcPCons tcPRef.

(********** TypeCheckPattern is a (partial) function **********)
Lemma TypeCheckPattern_deterministic: (E:envt)(P:pattern)(E1:envt)(t1:type)
 (TypeCheckPattern E P E1 t1) -> (E2:envt)(t2:type)
 (TypeCheckPattern E P E2 t2) ->
 ((E1=E2)/\(t1=t2)).
Proof.
Intros E P E1 t1 TCP; Induction TCP; Intros E3 t3 TCP2; Inversion TCP2; Auto.
 (* POpCons *)
 LApply (HrecTCP1 E4 t0); Auto; Intros (E14,t00).
 Rewrite E14 in HrecTCP0.
 LApply (HrecTCP0 E3 t4); Auto; Intros (E23,t14).
 Rewrite t00 in H.
 Cut((FunType t1 t2)=(FunType t4 t3)); [Intros EQF| EAuto].
 Inversion EQF; Split; Auto.
 (* PCons *)
 LApply (HrecTCP1 E4 t0); Auto; Intros (E14,t00).
 Rewrite E14 in HrecTCP0.
 LApply (HrecTCP0 E3 t4); Auto; Intros (E23,t14).
 Rewrite t00; Rewrite t14; Split; Auto.
 (* PRef *)
 LApply (HrecTCP E3 t0); Auto; Intros (E13,t00).
 Rewrite t00; Split; Auto.
Qed.
Hints Resolve TypeCheckPattern_deterministic.

(********** TypeCheckPattern ??? **********)
Lemma TypeCheckPattern_EqFun: (E1:envt)(P:pattern)(E10:envt)(t:type)
 (TypeCheckPattern E1 P E10 t) -> (E2:envt)
 (Eq E1 E2) ->
 (EX E20:envt | ((TypeCheckPattern E2 P E20 t) /\ (Eq E10 E20))).
Proof.
Intros E1 P E10 t TCP; Induction TCP; EAuto.
 (* POpCons *)
 Intros.
 LApply(HrecTCP1 E0); Auto; Intros (E5,(TCP5,EQ5)).
 LApply(HrecTCP0 E5); Auto; Intros (E6,(TCP6,EQ6)); EAuto.
 (* Pvar *)
 Intros.
 Exists (extend_envt E2 x t); Split; Auto.
 (* CARE Put that in Partial Function module *)
 Unfold Eq extend_envt Extend; Auto.
 Intros x0; Switch '(eqvar x x0); Auto.
 (* PCons *)
 Intros.
 LApply(HrecTCP1 E0); Auto; Intros (E5,(TCP5,EQ5)).
 LApply(HrecTCP0 E5); Auto; Intros (E6,(TCP6,EQ6)); EAuto.
 (* PRef *)
 Intros.
 LApply(HrecTCP E2); Auto; Intros (E5,(TCP5,EQ5)).
 Exists E5; Split; Auto.
Qed.
Hints Resolve TypeCheckPattern_EqFun.

(*
 * Type Check Expression
 *)
Inductive TypeCheckExpr : envt -> expr -> type -> Prop := 
   tcOp: (E:envt)(o:boperator)(t:type)
	(TypeCheckExpr E (Op (o,t)) t)
 | tcVar: (E:envt)(x:var)(t:type)
	(E x)=(Some type t) ->
	(TypeCheckExpr E (Var x) t)
 | tcAbs: (E,E1:envt)(P:pattern)(t1:type)
	(TypeCheckPattern E P E1 t1) -> (A:expr)(t2:type)
	(TypeCheckExpr E1 A t2) ->
	(TypeCheckExpr E (Abs P A) (FunType t1 t2))
 | tcApp: (E:envt)(F:expr)(t:type)
	(TypeCheckExpr E F t) -> (t1,t2:type)
        (NormalizeFunType t (FunType t1 t2)) -> (A:expr)
	(TypeCheckExpr E A t1) ->
	(TypeCheckExpr E (App F A) t2)
 | tcCons: (E:envt)(A:expr)(t1:type)
	(TypeCheckExpr E A t1) -> (B:expr)(t2:type)
	(TypeCheckExpr E B t2) ->
	(TypeCheckExpr E (Cons A B) (ProdType t1 t2))
 | tcRef: (E:envt)(A:expr)(t:type)
	(TypeCheckExpr E A t) ->
	(TypeCheckExpr E (Ref A) (RefType t))
 | tcDeref: (E:envt)(A:expr)(t:type)
	(TypeCheckExpr E A (RefType t)) ->
	(TypeCheckExpr E (Deref A) t)
 | tcAssign: (E:envt)(A:expr)(t1:type)
	(TypeCheckExpr E A (RefType t1)) -> (B:expr)(t2:type)
	(TypeCheckExpr E B t1) ->
	(TypeCheckExpr E (Assign A B) t1).
Hints Resolve tcOp tcVar tcAbs tcApp tcCons tcRef tcDeref tcAssign.

(********** TypeCheckExpr is a (partial) function **********)
Lemma TypeCheckExpr_deterministic: (E:envt)(A:expr)(t1:type)
 (TypeCheckExpr E A t1) -> (t2:type)
 (TypeCheckExpr E A t2) ->
 (t1=t2).
Proof.
Intros E A t1 TE; Induction TE; Intros t3 TE2; Inversion TE2; EAuto.
 (* Var *)
 Rewrite H in H2; Inversion H2; Auto.
 (* Abs *)
 Cut((E1=E2)/\(t1=t0)); [Intros (E12,t10) | EAuto].
 Rewrite E12 in HrecTE.
 LApply (HrecTE t4); Auto; Intros t24.
 Rewrite t10; Rewrite t24; Auto.
 (* App *)
 LApply (HrecTE1 t0); Auto; Intros t00.
 Rewrite t00 in H.
 Cut((FunType t1 t2)=(FunType t4 t3)); [Intros EQF| EAuto].
 Inversion EQF; Auto.
 (* Cons *)
 LApply (HrecTE1 t0); Auto; Intros t10.
 LApply (HrecTE0 t4); Auto; Intros t24.
 Rewrite t10; Rewrite t24; Auto.
 (* Ref *)
 LApply (HrecTE t0); Auto; Intros t10.
 Rewrite t10; Auto.
 (* Assign *)
 LApply (HrecTE (RefType t3)); Auto; Intros EQR.
 Inversion EQR; Auto.
Qed.
Hints Resolve TypeCheckExpr_deterministic.

(********** TypeCheckExpr ??? **********)
Lemma TypeCheckExpr_EqFun: (E1:envt)(A:expr)(t:type)
 (TypeCheckExpr E1 A t) -> (E2:envt)
 (Eq E1 E2) ->
 (TypeCheckExpr E2 A t).
Proof.
Intros E1 A t TE; Induction TE; Intros E2 EQF; EAuto.
 Unfold Eq in EQF; Cut((E x)=(E2 x)); [Intros EQ | EAuto].
 Rewrite EQ in H; Auto.
Cut(EX E3:envt | ((TypeCheckPattern E2 P E3 t1) /\ (Eq E1 E3)));
  [Intros (E3,(TCP3,EQ3)) | EAuto].
LApply(HrecTE E3); Auto; Intros TCE; EAuto.
Qed.
Hints Resolve TypeCheckExpr_EqFun.

(*
 * Type of Values (lemmas are with AbstractEnv)
 *)
Mutual Inductive TypeOf : storet -> value -> type -> Prop :=
 | toOpNil: (S:storet)(o:boperator)(t:type)
	(TypeOf S (VOp (o,t) (nil ?)) t)
 | toOpCons: (S:storet)(op:operator)(l:values)(t:type)
	(TypeOf S (VOp op l) t) -> (t1,t2:type)
	(NormalizeFunType t (FunType t1 t2)) -> (v:value)
	(TypeOf S v t1) ->
	(TypeOf S (VOp op (cons v l)) t2)
 | toClosure: (S:storet)(e:env)(E:envt)
	(AbstractEnv S e E) -> (P:pattern)(B:expr)(t1,t2:type)
	(TypeCheckExpr E (Abs P B) (FunType t1 t2)) ->
	(TypeOf S (Closure P B e) (FunType t1 t2))
 | toPair: (S:storet)(Av:value)(t1:type)
	(TypeOf S Av t1) -> (Bv:value)(t2:type)
	(TypeOf S Bv t2) ->
	(TypeOf S (Pair Av Bv) (ProdType t1 t2))
 | toLoc: (S:storet)(i:loc)(t:type)
	(S i)=(Some type (RefType t)) ->
	(TypeOf S (Loc i) (RefType t))
 | toWrong: (S:storet)(e:env)(E:envt)
	(AbstractEnv S e E) -> (P:pattern)(B:expr)(t1,t2:type)
	(TypeCheckExpr E (Abs P B) (FunType t1 t2)) -> (v:value)
	(TypeOf S v t1) ->
	(TypeOf S (Wrong P v B e) t2)

(*
 * Abstraction of an environment
 *)
with    AbstractEnv: storet -> env -> envt -> Prop :=
 | aeInit: (S:storet)
	(AbstractEnv S env_init envt_init)
 | aeExtend: (S:storet)(e:env)(E:envt)
	(AbstractEnv S e E) -> (v:value)(t:type)
	(TypeOf S v t) -> (x:var)
	(AbstractEnv S (extend_env e x v) (extend_envt E x t)).
Hints Resolve toOpNil toOpCons toClosure toPair toLoc toWrong.
Hints Resolve aeInit aeExtend.

Scheme nTypeOf_ind := Minimality for TypeOf Sort Prop
  with nAbstractEnv_ind := Minimality for AbstractEnv Sort Prop.

(*
 * Lemmas for TypeOf
 *)

(********** Same type value for growing abstract store **********)
Definition tltTypeOf : storet -> value -> type -> Prop :=
 [S:storet][v:value][t:type]
 ((S2:storet)
  (Lower S S2) ->
  (TypeOf S2 v t) ).

Definition tltAbstractEnv : storet -> env -> envt -> Prop :=
 [S:storet][e:env][E:envt]
 ((S2:storet)
  (Lower S S2) ->
  (AbstractEnv S2 e E) ).

Lemma TypeOf_Lower_TypeOf: (S:storet)(v:value)(t:type)
 (TypeOf S v t) -> (S2:storet)
 (Lower S S2) ->
 (TypeOf S2 v t).
Proof.
Intros S v t TO;
Elim TO using nTypeOf_ind with tltTypeOf tltAbstractEnv S v t;
Clear TO S v t; Unfold tltTypeOf tltAbstractEnv; EAuto.
Qed.
Hints Resolve TypeOf_Lower_TypeOf.

(********** Same type on clean extension **********)
Lemma GrowingTypeOf: (S:storet)(v:value)(t:type)
 (TypeOf S v t) -> (i:loc)
 (S i)=(None type) -> (t':type)
 (TypeOf (extend_storet S i (RefType t')) v t).
Intros.
Cut(Lower S (extend_storet S i (RefType t')));
 [EAuto | Apply (Fresh_LowerExtend beq_nat_eq H0)].
Qed.
Hints Resolve GrowingTypeOf.

(* Specific Lemma for POp/Vop construction *)
Lemma OpTypeAux: (n:nat)(lp:patterns)(l:values)
  (n=(length lp)) -> 
  (n=(length l)) -> (E,E2:envt)(op:operator)(t:type)
 (TypeCheckPattern E (POp op lp) E2 t) -> (S:storet)(t3:type)
 (TypeOf S (VOp op l) t3) ->
 (t=t3).
Proof.
Intros n; Induction n; Intros lp l.
 Case lp; Case l; Unfold length; Intros; Inversion H; Inversion H0.
 Generalize H1 H2; Case op; Intros b t0 TCP; Inversion TCP.
 Intros TO; Inversion TO; Auto.
Case lp; Clear lp; [(Intros X; Inversion X) | Intros v lp SLP].
Case l; Clear l; [(Intros X; Inversion X) | Intros v0 l SL].
Cut(n=(length lp)); [Intros LP | Auto]; Clear SLP.
Cut(n=(length l)); [Intros L | Auto]; Clear SL.
Intros E E2 op t TCP; Inversion TCP; Intros S t3 VC; Inversion VC.
LApply (Hrecn lp l LP L E E1 op t0 H3 S t4); Auto; Clear Hrecn; Intros t04.
Rewrite t04 in H6.
Cut((FunType t1 t)=(FunType t5 t3)); [Intros EQF | EAuto].
Inversion EQF; Auto.
Qed.

Lemma SameOpType: (lp:patterns)(l:values)
 ((length lp)=(length l)) -> (E,E2:envt)(op:operator)(t:type)
 (TypeCheckPattern E (POp op lp) E2 t) -> (S:storet)(t3:type)
 (TypeOf S (VOp op l) t3) ->
 (t=t3).
Proof.
Intros lp l EQ.
Cut(EX n:nat | n=(length lp)); [Intros (n,EQn) | Exists (length lp)]; Auto.
Rewrite <- EQn in EQ.
Apply (OpTypeAux EQn EQ).
Qed.

(*
 * Lemmas for AbstractEnv
 *)
(********** What we need for self abstracted environment **********)
Lemma WellAbstractedEnv: (S:storet)(e:env)(E:envt)
 (AbstractEnv S e E) -> (x:var)(v:value)
 (e x)=(Some value v) -> (t:type)
 (E x)=(Some type t) ->
 (TypeOf S v t).
Proof.
Intros S e E AE; Induction AE; Intros x0 v0.
 Intros BAD; Unfold env_init in BAD; Inversion BAD.
Unfold extend_env extend_envt Extend; Switch '(eqvar x x0); EAuto.
Intros EQv t0 EQt; Generalize H; Inversion EQv; Inversion EQt; Auto.
Qed.
Hints Resolve WellAbstractedEnv.

(********** Same Abstract Environment for growing abstract store **********)
Lemma AbstractEnv_Lower_AbstractEnv: (e:env)(E:envt)(S:storet)
 (AbstractEnv S e E) -> (S2:storet)
 (Lower S S2) ->
 (AbstractEnv S2 e E).
Proof.
Intros e E S AS; Induction AS; EAuto.
Qed.
Hints Resolve AbstractEnv_Lower_AbstractEnv.

(*
 * Abstraction of a store
 *)
Definition AbstractStore: storet -> store -> storet -> Prop :=
 [S1:storet][s:store][S2:storet]
  ( ((i:loc)(v:value) (s i)=(Some value v) ->
	(EX t:type | ((S2 i)=(Some type (RefType t)) /\
                      (TypeOf S1 v t) )))
 /\ ((i:loc) (s i)=(None value) -> (S2 i)=(None type)) ).

(********** What we need for self abstracted store **********)
Lemma WellAbstractedStore: (S:storet)(s:store)
 (AbstractStore S s S) -> (i:loc)(v:value)
 (s i)=(Some value v) -> (t:type)
 (S i)=(Some type (RefType t)) ->
 (TypeOf S v t).
Proof.
Unfold AbstractStore; Intros S s (RG,RN) i v GET t TGET.
LApply (RG i v); Auto; Intros (tt,(TGET2,TO2)).
Rewrite TGET in TGET2; Inversion TGET2; Auto.
Qed.
Hints Resolve WellAbstractedStore.

(********** Step for Well Abstracted Store **********)
Lemma AbstractStore_TypeOf_AbstractStore: (S:storet)(s:store)
 (AbstractStore S s S) -> (v:value)(t:type)
 (TypeOf S v t) -> (i:loc)
 (TypeOf S (Loc i) (RefType t)) ->
 (AbstractStore S (extend_store s i v) S).
Proof.
Unfold AbstractStore; Intros S s (OK,NONE) v t TO1 i TO2; Split.
 Intros i1 v0; Unfold extend_store; Unfold Extend.
 Switch '(beq_nat i i1); Auto.
 Cut(i=i1); [Intros EQ | Apply beq_nat_eq]; Auto; Rewrite <- EQ.
 Intros EQv; Inversion EQv; Rewrite <- H0.
 Exists t; Split; Inversion TO2; Auto.
Intros i1; Unfold extend_store; Unfold Extend.
Switch '(beq_nat i i1); Auto.
Intros Bad; Inversion Bad.
Qed.
Hints Resolve AbstractStore_TypeOf_AbstractStore.

(********** Step Coherence on sbrk (Part1) **********)
Lemma AbstractStore_sbrk_Lower: (S:storet)(s:store)
 (AbstractStore S s S) -> (i:loc)
 i=(sbrk s) -> (t:type)
 (Lower S (extend_storet S i (RefType t))).
Proof.
Unfold AbstractStore; Intros S s (OK,NONE) i SBRK t.
Unfold Lower; Intros.
Unfold extend_storet; Unfold Extend.
Switch '(beq_nat i x); Auto.
Cut((s i)=(None value)); [Intros FRESH | Apply sbrk_valid]; Auto.
LApply (NONE i); Auto; Intros TNONE.
Cut(i=x); [Intros EQ | Apply beq_nat_eq]; Auto.
Rewrite <- EQ in H. Rewrite TNONE in H; Inversion H.
Qed.
Hints Resolve AbstractStore_sbrk_Lower.

(********** Step Coherence on sbrk (Part2) **********)
Lemma AbstractStore_sbrk_AbstractStore: (S:storet)(s:store)
 (AbstractStore S s S) -> (i:loc)
 i=(sbrk s) -> (a:value)(t:type)
 (TypeOf S a t) ->
 (AbstractStore (extend_storet S i (RefType t)) (extend_store s i a)
                (extend_storet S i (RefType t))).
Unfold AbstractStore; Intros S s (RG,RN) i SBRK a t TO.
Cut((s i)=(None value)); [Intros FRESH | Apply sbrk_valid]; Auto.
LApply (RN i); Auto; Intros TNONE.
Split.
 Intros i1 v; Unfold extend_store Extend.
 Switch '(beq_nat i i1); Auto.
  Cut(i=i1); [Intros EQ | Apply beq_nat_eq]; Auto; Rewrite <- EQ.
  Intros EQv; Inversion EQv; Rewrite <- H0.
  Exists t; Split; Auto.
   Unfold extend_storet Extend.
   Cut(true=(beq_nat i i)); [Intros EQi | Apply beq_nat_refl].
   Rewrite <- EQi; Auto.
 Intros GET.
 LApply (RG i1 v); Auto; Intros (tv,(TGETi1,TO1)).
 Exists tv; Split; Auto.
  Unfold extend_storet Extend.
  Rewrite <- RES; Auto.
Intros i0; Unfold extend_store extend_storet Extend.
Switch '(beq_nat i i0); [(Intros BAD; Inversion BAD) | Auto].
Qed.
Hints Resolve AbstractStore_sbrk_AbstractStore.

(*
 * Full Abstraction: Abstract an environment AND a store
 *)
Definition FixAbstract: env -> store -> envt -> storet -> Prop :=
 [e:env][s:store][E:envt][S:storet]
 ((AbstractEnv S e E) /\ (AbstractStore S s S)).

(********** first projection **********)
Lemma FixAbstract_AbstractEnv: (e:env)(s:store)(E:envt)(S:storet)
 (FixAbstract e s E S) ->
 (AbstractEnv S e E).
Proof.
Unfold FixAbstract; Intros e s E S (AE,AS); Auto.
Qed.
Hints Resolve FixAbstract_AbstractEnv.

(********** second projection **********)
Lemma FixAbstract_AbstractStore: (e:env)(s:store)(E:envt)(S:storet)
 (FixAbstract e s E S) ->
 (AbstractStore S s S).
Proof.
Unfold FixAbstract; Intros e s E S (AE,AS); Auto.
Qed.
Hints Resolve FixAbstract_AbstractStore.

(********** Increasing the first projection **********)
Lemma FixAbstract_AbstractEnv_FixAbstract: (e:env)(s:store)(E:envt)(S:storet)
 (FixAbstract e s E S) -> (e2:env)(E2:envt)
 (AbstractEnv S e2 E2) ->
 (FixAbstract e2 s E2 S).
Proof.
Unfold FixAbstract; Intros e s E S (AE,AS); Auto.
Qed.
Hints Resolve FixAbstract_AbstractEnv_FixAbstract.

(********** Increasing the second projection **********)
Lemma FixAbstract_Lower_AbstractStore_FixAbstract:
       (e:env)(s:store)(E:envt)(S:storet)
 (FixAbstract e s E S) -> (S2:storet)
 (Lower S S2) -> (s2:store)
 (AbstractStore S2 s2 S2) -> 
 (FixAbstract e s2 E S2).
Proof.
Unfold FixAbstract; Intros e s E S (AE,AS); EAuto.
Qed.
Hints Resolve FixAbstract_Lower_AbstractStore_FixAbstract.

(********** Full Abstraction have well abstracted store **********)
Lemma FixAbstract_WellAbstractedStore:(e:env)(s:store)(E:envt)(S:storet)
 (FixAbstract e s E S) -> (i:loc)(v:value)
 (s i)=(Some value v) -> (t:type)
 (S i)=(Some type (RefType t)) ->
 (TypeOf S v t).
Proof.
Unfold FixAbstract; Intros e s E S (AE,AS); EAuto.
Qed.
Hints Resolve FixAbstract_WellAbstractedStore.

(********** Full Abstraction have well abstracted environment **********)
Lemma FixAbstract_WellAbstractedEnv:(e:env)(s:store)(E:envt)(S:storet)
 (FixAbstract e s E S) -> (x:var)(v:value)
 (e x)=(Some value v) -> (t:type)
 (E x)=(Some type t) ->
 (TypeOf S v t).
Proof.
Unfold FixAbstract; Intros e s E S (AE,AS); EAuto.
Qed.
Hints Resolve FixAbstract_WellAbstractedEnv.

(*
 * Coherence between two couples of a store and its abstraction
 *)
Definition Coherent: store -> storet -> store -> storet -> Prop :=
 [s:store][S:storet][s2:store][S2:storet]
 ((Lower S S2)  /\ (AbstractStore S2 s2 S2)).

(********** Reflexivity via self abstracted store **********)
Lemma Coherent_refl: (s:store)(S:storet)
 (AbstractStore S s S) ->
 (Coherent s S s S).
Proof.
Unfold Coherent; Intros; Auto.
Qed.
Hints Resolve Coherent_refl.

(********** Reflexivity via full abstraction **********)
Lemma FixAbstract_Coherent_refl: (e:env)(s:store)(E:envt)(S:storet)
 (FixAbstract e s E S) ->
 (Coherent s S s S).
Proof.
Unfold Coherent; EAuto.
Qed.
Hints Resolve FixAbstract_Coherent_refl.

(********** Transitivity **********)
Lemma Coherent_trans: (s:store)(S:storet)(s1:store)(S1:storet)
 (Coherent s S s1 S1) -> (s2:store)(S2:storet)
 (Coherent s1 S1 s2 S2) ->
 (Coherent s S s2 S2).
Proof.
Unfold Coherent; Intros s S s1 S1 (LE1,AS2) s2 S2 (LE2,AS3).
EAuto.
Qed.
Hints Resolve Coherent_trans.

(********** The second component must self abstracted **********)
Lemma Coherent_AbstractStore: (s:store)(S:storet)(s1:store)(S1:storet)
 (Coherent s S s1 S1) ->
 (AbstractStore S1 s1 S1).
Proof.
Unfold Coherent. Intros s S s1 S1 (LE,AS2); Auto.
Qed.
Hints Resolve Coherent_AbstractStore.

(********** Typed values must follow **********)
Lemma Coherent_TypeOf: (s1:store)(S1:storet)(s2:store)(S2:storet)
 (Coherent s1 S1 s2 S2) -> (v:value)(t:type)
 (TypeOf S1 v t) ->
 (TypeOf S2 v t).
Proof.
Unfold Coherent; Intros s1 S1 s2 S2 (LE,X) v t TO; EAuto.
Qed.
Hints Resolve Coherent_TypeOf.

(********** Type environment must follow **********)
Lemma Coherent_AbstractEnv: (s1:store)(S1:storet)(s2:store)(S2:storet)
 (Coherent s1 S1 s2 S2) -> (e:env)(E:envt)
 (AbstractEnv S1 e E) ->
 (AbstractEnv S2 e E).
Proof.
Unfold Coherent; Intros s1 S1 s2 S2 (LE,X) v t AE; EAuto.
Qed.
Hints Resolve Coherent_AbstractEnv.

(********** Full Abstraction must follow **********)
Lemma Coherent_FixAbstract: (s1:store)(S1:storet)(s2:store)(S2:storet)
 (Coherent s1 S1 s2 S2) -> (e:env)(E:envt)
 (FixAbstract e s1 E S1) ->
 (FixAbstract e s2 E S2).
Proof.
Unfold Coherent; Intros s1 S1 s2 S2 (LE,AS2) e E FX; EAuto.
Qed.
Hints Resolve Coherent_FixAbstract.

(********** Step Coherence on sbrk **********)
Lemma sbrk_Coherent: (S:storet)(s:store)
 (AbstractStore S s S) -> (i:loc)
 i=(sbrk s) -> (a:value)(t:type)
 (TypeOf S a t) ->
 (Coherent s S (extend_store s i a) (extend_storet S i (RefType t))).
Proof.
Unfold Coherent; EAuto.
Qed.
Hints Resolve sbrk_Coherent.

(********** Step Coherence on store extension **********)
Lemma AbstractStore_TypeOf_Coherent: (S:storet)(s:store)
 (AbstractStore S s S) -> (v:value)(t:type)
 (TypeOf S v t) -> (i:loc)
 (TypeOf S (Loc i) (RefType t)) ->
 (Coherent s S (extend_store s i v) S).
Proof.
Unfold Coherent; EAuto.
Qed.
Hints Resolve AbstractStore_TypeOf_Coherent.

(* 
 *
 * Subject Reduction
 *
 *)

(*
 * Subject Reduction for the Matching Algorithm
 *)
Lemma open_subject_reduction_match: (P:pattern)(v:value)(s:store)(e,e':env)
 (match P v s e e') -> (E:envt)(S:storet)
 (FixAbstract e s E S) -> (E1:envt)(t1:type)
 (TypeCheckPattern E P E1 t1) -> 
 (TypeOf S v t1) ->
 (AbstractEnv S e' E1).
Proof.
Intros P v s e e' M; Elim M; Clear M P v s e e'.
 (* POpNil*)
 Unfold FixAbstract; Intros op s e E S (AE,AS) E1 t1 TCP TO; Inversion TCP.
 Rewrite <- H0; Auto.
 (* POpCons *)
 Intros op lp l s e e' Ml rMl p v r MP rMP E S FX E1 t1 TCP; Inversion TCP.
 Intros TO; Inversion TO.
 Cut((length lp)=(length l)); [Intros EQ | Apply (Match_Op_Length Ml)].
 Cut(t=t3); [Intro EQT | Apply (SameOpType EQ H3 H12)].
 Rewrite <- EQT in H12; Rewrite EQT in H6.
 LApply (rMl E S FX E2 t H3); Auto; Clear rMl; Intros AS1.
 Cut((FunType t0 t1)=(FunType t4 t1)); [Intros EQF | EAuto].
 Inversion EQF; Rewrite <- H17 in H15.
 Cut(FixAbstract e' s E2 S); [Intros FX2 | Auto]; EAuto.
 (* PVar *)
 Intros; Inversion H0; EAuto.
 (* PCons *)
 Intros p1 a1 s e e' M1 rM1 p2 a2 r M2 rM2 E S FX E1 t1 TCP; Inversion TCP.
 Intros TO; Inversion TO.
 LApply (rM1 E S FX E2 t0 H2); Auto; Clear rM1; Intros AE1.
 Cut(FixAbstract e' s E2 S); EAuto.
 (* Pref *)
 Intros i s v GET x e r M rM E S FX E1 t1 TCP; Inversion TCP.
 Intros TO; Inversion TO.
 Cut(TypeOf S v t); [Intros TOv | EAuto]; EAuto.
Qed.
Hints Resolve open_subject_reduction_match.

(*
 * Subject Reduction for Open Expression
 *)
Definition opsEVAL : expr -> env -> store -> value -> store -> Prop :=
 [A:expr][e:env][s:store][v:value][s2:store]
 ((E:envt)(S:storet)
  (FixAbstract e s E S) -> (t:type)
  (TypeCheckExpr E A t) ->
  (EX S2:storet | ((Coherent s S s2 S2) /\ (TypeOf S2 v t))) ).

Definition opsCALL : value -> value -> store -> value -> store -> Prop :=
 [f:value][a:value][s:store][v:value][s2:store]
 ((S:storet)(t:type)
  (AbstractStore S s S) ->
  (TypeOf S f t) -> (t1,t2:type)
  (NormalizeFunType t (FunType t1 t2)) ->
  (TypeOf S a t1) ->
  (EX S2:storet | ((Coherent s S s2 S2) /\ (TypeOf S2 v t2))) ).

Tactic Definition Got1 := Intros (S1,(CO1,TOf1)).
Tactic Definition Got2 := Intros (S2,(CO2,TOf2)).
Tactic Definition Got3 := Intros (S3,(CO3,TOf3)).

Lemma open_subject_reduction: (A:expr)(e:env)(s:store)(v:value)(s2:store)
 (eval A e s v s2) -> (E:envt)(S:storet)
 (FixAbstract e s E S) -> (t:type)
 (TypeCheckExpr E A t) ->
 (EX S2:storet | ((Coherent s S s2 S2) /\ (TypeOf S2 v t))).
Proof.
Intros A e s1 v s2 EV;
Elim EV using neval_ind with opsEVAL opsCALL A e s1 v s2;
Clear EV A e s1 v s2; Unfold opsEVAL; Unfold opsCALL; EAuto.
 (* Op *)
 Intros op e s E S ABS t TE; Inversion TE; Exists S; Split; EAuto.
 (* Var *)
 Intros x e v GET s E S ABS t TE; Inversion TE; Exists S; Split; EAuto.
 (* Abs *)
 Intros P B e s E S ABS t TE; Inversion TE; Exists S; Split; EAuto.
 (* App *)
 Intros F e s f s1 EF rF A a s2 EA rA v s3 CL rCL E S ABS t TE; Inversion TE;
 LApply (rF E S ABS t0); Auto; Clear EF rF; Got1.
 Cut(FixAbstract e s1 E S1); [Intros ABS1 | EAuto].
 LApply (rA E S1 ABS1 t1); Auto; Clear EA rA; Got2.
 Cut(FixAbstract e s2 E S2); [Intros ABS2 | EAuto].
 Cut(TypeOf S2 f t0); [Intros Tf | EAuto].
 Cut(AbstractStore S2 s2 S2); [Intros AS2 | EAuto].
 LApply(rCL S2 t0 AS2 Tf t1 t H3); Auto; Got3.
 Exists S3; Split; Auto; EAuto.
 (* Cons *)
 Intros A e s a s1 EA rA B b s2 EB rB E S ABS t TE; Inversion TE.
 LApply (rA E S ABS t1); Auto; Clear rA; Got1.
 Cut(FixAbstract e s1 E S1); [Intros ABS1 | EAuto].
 LApply (rB E S1 ABS1 t2); Auto; Clear rB; Got2.
 Exists S2; Split; [EAuto | Auto]; Apply toPair; Auto; EAuto.
 (* Ref *)
 Intros A e s a s1 EA rA i SBRK E S ABS t TE; Inversion TE.
 LApply (rA E S ABS t0); Auto; Clear rA; Got1.
 Cut(FixAbstract e s1 E S1); [Intros ABS1 | EAuto].
 Cut(AbstractStore S1 s1 S1); [Intros AS1 | EAuto].
 Cut(Coherent s1 S1 (extend_store s1 i a) (extend_storet S1 i (RefType t0)));
  [Intros CO2 | Auto].
 Exists (extend_storet S1 i (RefType t0)); Split; [EAuto | Apply toLoc].
 Unfold extend_storet; Unfold Extend.
 Cut(true=(beq_nat i i)); [Intros EQ | Apply beq_nat_refl]; Rewrite <- EQ; Auto.
 (* Deref *)
 Intros A e s i s1 EA rA v GET E S ABS t TE; Inversion TE.
 LApply (rA E S ABS (RefType t)); Auto; Clear rA; Got1.
 Exists S1; Split; Auto; Inversion TOf1.
 Cut(FixAbstract e s1 E S1); [Intros ABS1 | EAuto].
 EAuto.
 (* Assign *)
 Intros A e s i v s1 EA rA B s2 EB rB old GET E S ABS t TE; Inversion TE.
 LApply (rA E S ABS (RefType t)); Auto; Clear rA; Got1.
 Cut(FixAbstract e s1 E S1); [Intros ABS1 | EAuto].
 LApply (rB E S1 ABS1 t); Auto; Clear rB; Got2.
 Exists S2; Split; Auto.
 Cut(TypeOf S2 (Loc i) (RefType t)); [Intros T2 | EAuto].
 Cut(FixAbstract e s2 E S2); [Intros ABS2 | EAuto].
 Cut(AbstractStore S2 s2 S2); [Intros FA | EAuto].
 Cut(Coherent s2 S2 (extend_store s2 i v) S2); [Intros CO3 | EAuto]; EAuto.
 (* CALL *)
 (* Op *)
 (* Closure OK *)
 Intros P v s e e' M B r s1 EB rB S t AS TO; Inversion TO.
 Inversion H5; Intros t4 t5 NFT TO2; Inversion NFT.
 Rewrite <- H16 in TO2.
 Cut(FixAbstract e s E S); [Intros | (Unfold FixAbstract; Auto)].
 Cut(AbstractEnv S e' E1); [Intros ABSENV2 |
   Apply (open_subject_reduction_match M H13 H10 TO2) ].
 Cut(FixAbstract e' s E1 S); [Intros ABS2 | (Unfold FixAbstract; Auto)].
 LApply (rB E1 S ABS2 t2); Auto; Clear rB; Got1.
 Exists S1; Split; [Auto | Rewrite <- H17]; Auto.
 (* CLosure KO *)
 Intros P v s e NOT B S t AS TO; Inversion TO.
 Inversion H5; Intros t4 t5 NFT TO2; Inversion NFT.
 Rewrite <- H17.
 Exists S; Split; [EAuto | Apply (toWrong H4 H5)].
 Rewrite H16; Auto.
(* Pair *)
 Intros a v s r1 s1 Ca rCa b r2 s2 Cb rCb S t AS TO; Inversion TO.
 Intros t3 t4 NFT TO2.
 Inversion NFT.
 LApply (rCa S t1 AS H2 t3 t6 H8); Auto; Clear rCa; Got1.
 Cut(TypeOf S1 b t2); [Intros TOB | EAuto].
 Cut(TypeOf S1 v t3); [Intros TO3 | EAuto].
 Cut(AbstractStore S1 s1 S1); [Intros AS1 | EAuto].
 LApply (rCb S1 t2 AS1 TOB t3 t8 H10); Auto; Clear rCb; Got2.
 Exists S2; Split; [EAuto | Auto]; Apply toPair; Auto; EAuto.
Qed.
Hints Resolve open_subject_reduction.

(*
 * Finally the Subject Reduction Theorem
 *)
Lemma subject_reduction_init:
 (FixAbstract env_init store_init envt_init storet_init).
Proof.
Unfold FixAbstract; Unfold AbstractStore; Split; Auto; Split; Auto.
Intros; Unfold store_init in H; Inversion H.
Qed.
Hints Resolve subject_reduction_init.

Theorem subject_reduction: (A:expr)(v:value)(s2:store)
 (eval A env_init store_init v s2) -> (t:type)
 (TypeCheckExpr envt_init A t) ->
 (EX S2:storet | ((Coherent store_init storet_init s2 S2) /\ (TypeOf S2 v t))).
Proof.
EAuto.
Qed.

End TypedImpRho.
